library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.ALL;

-- Peripheral controller

-- Need to provide:
-- 	UART
--    X"000" - UART_IO
--    	On Write: Sends lower 8 bits.  Warning: will block until previous write is complete!
--    	On Read:  Bit1 10: txint, bit 9: rxint, Bit 8: txready.  Bits 7-0: received data.
--		X"002" - UART_CLKDIV
--			R/W - Gets/Sets current clock divisor for baud rate generator

-- 	Timers?
--		SD access?

entity peripheral_controller is
  port (
		clk : in std_logic;
		reset : in std_logic;

		reg_addr_in : in std_logic_vector(11 downto 0); -- from host CPU
		reg_data_in: in std_logic_vector(15 downto 0);
		reg_data_out: out std_logic_vector(15 downto 0);
		reg_rw : in std_logic;
		reg_uds : in std_logic;
		reg_lds : in std_logic;
		reg_dtack : out std_logic;	-- Needed for char ram access.
		reg_req : in std_logic;
		uart_int : out std_logic;

		uart_rxd : in std_logic;
		uart_txd : out std_logic
	);
end entity;
	
architecture rtl of peripheral_controller is

signal ser_txdata : std_logic_vector(7 downto 0);
signal ser_txready : std_logic;
signal ser_txgo : std_logic :='0';
signal ser_rxdata : std_logic_vector(7 downto 0);
signal ser_rxint : std_logic;
signal ser_txint : std_logic;
signal ser_ints : std_logic_vector(1 downto 0);
signal ser_clock_divisor : unsigned(15 downto 0) := X"16E3";	-- 19200 baud @ 112.5MHz

begin
	-- Handle CPU access to hardware registers

	myuart : entity work.simple_uart
		port map(
			clk => clk,
			reset => reset,
			txdata => ser_txdata,
			txready => ser_txready,
			txgo => ser_txgo,
			rxdata => ser_rxdata,
			rxint => ser_rxint,
			txint => ser_txint,
			clock_divisor => ser_clock_divisor,
			rxd => uart_rxd,
			txd => uart_txd
		);
	
	process(clk,reset)
	begin
		if reset='0' then
			uart_int<='0';
			ser_ints<="00";
		elsif rising_edge(clk) then
			reg_dtack<='1';
			ser_txgo<='0';
			
			-- Handle serial interrupts.  We map both Rx and Tx interrupts onto a single interrupt
			-- signal, but keep track of which has occured.
			-- When the CPU reads the register these are cleared.
			uart_int<=ser_rxint or ser_txint;
			if ser_rxint='1' then
				ser_ints(0)<='1';
			end if;
			if ser_txint='1' then
				ser_ints(1)<='1';
			end if;
	
			if reg_req='1' then
				case reg_addr_in is
					when X"000" =>
						if reg_rw='0' and ser_txready='1' then  -- write
							ser_txdata<=reg_data_in(7 downto 0);
							ser_txgo<='1';
							reg_dtack<='0';
						else -- read
							reg_data_out<="00000"&ser_ints&ser_txready&ser_rxdata(7 downto 0);
							reg_dtack<='0';
							ser_ints<="00"; -- Clear int flags
						end if;
					when X"002" =>
						if reg_rw='0' then
							ser_clock_divisor<=unsigned(reg_data_in);
						else -- read
							reg_data_out<=std_logic_vector(ser_clock_divisor);
						end if;
						reg_dtack<='0';						
					when others =>
						reg_data_out<=X"0000";
						reg_dtack<='0';
				end case;
			end if;

		end if;
	end process;
		
end architecture;